/*
 * playitsam: Play, edit & save a TiVo video stream from stdin, file or MFS.
 *
 * $Revision: 1.38 $
 * 
 * (c) 2002, Warren Toomey wkt@tuhs.org.
 * Derived from vplay, written by tridge@samba.org, December 2000
 * 
 * Released under the Gnu General Public License version 2
 */

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <termios.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include "playitsam.h"
#ifdef TIVO
#include <sched.h>
#include "mfs.h"
#include "osd.h"

static int remote_fd;			/* fd for remote control device */
#endif

FILE *audio_f;				/* Non-TiVo audio pipe */
FILE *video_f;				/* Non-TiVo video pipe */
static int audio_fd;			/* fd for audio device */
static int video_fd;			/* fd for video device */

int verbose = 0;			/* Does user want verbose output */
int allow_playback = 1;			/* Are we going to play to devices? */
					/* If ==2, then only play, no editing */
int seeking_ok = 1;			/* Can we seek from stream file? */
int chunk, totalchunks;			/* Current and total chunks */

int exit_things=0;			/* Things to do at exit time */
struct termios oldtios, newtios;	/* Console settings for raw & cooked */
					/* modes */
unsigned char buf[CHUNK_SIZE];		/* Buffer for one chunk */

int cutstart[MAXCUTS], cutend[MAXCUTS];	/* Edited out chunk ranges */

int inputtype = 0;			/* Type of file input */

int do_osd = 1;				/* Do we want on-screen display? */

int osd_x=1, osd_y=15;			/* Co-ords for OSD string display */
int osd_dirty=0;			/* Is there any text on the screen? */
int osd_chunk=0;			/* If >0, this chunk # now osd */
char osd_string[80];			/* String to draw on the screen */


#ifdef TIVO
/* TODO: given a process name, find the matching pid.
 * Return the pid, or -1 if none.
 */
int pid_fromname(char *progname)
{
 DIR *d;
 FILE *cmdf;
 char *c, cmdline[80], cmdfile[1024];
 struct dirent *dp;

 d= opendir("/proc");				/* Open /proc directory */
 if (d==NULL) return(-1);

 while ((dp = readdir(d)) != NULL) {
   if (!isdigit(dp->d_name[0])) continue;	/* Skip non-digit filenames */

   snprintf(cmdfile, 1023, "/proc/%s/cmdline", dp->d_name);
   cmdf= fopen(cmdfile, "r");			/* Open /proc/NNN/cmdline */
   if (cmdf==NULL) continue;

   c=fgets(cmdline, 80, cmdf);			/* Read the command line */
   fclose(cmdf);				/* and close the file */
   if (c==NULL) continue;

   if (!strcmp(cmdline, progname)) {		/* If it matches what we have */
     (void)closedir(d);
     return(atoi(dp->d_name));			/* then return NNN as integer */
   }
 }
 (void)closedir(d);
 return(-1);					/* No match, return -1 */
}


/* Restart MyWorld. Thanks to D18C7DB for the tip. */
void start_myworld(void)
{
  int fd, pid;

  /* First run momstart */
  pid=fork();
  switch (pid) {
    case -1: return;
    case 0: execl("/tvbin/momstart", "/tvbin/momstart", "-m", NULL); exit(1);
    default: sleep(1);
  }

  /* Now run MyWorld */
  pid=fork();
  switch (pid) {
    case -1: return;
    case 0: fd=open("/dev/ttyS2", O_WRONLY);
	    if (fd<0) return;
	    close(1); dup2(fd, 1);
	    close(2); dup2(fd, 2);
	    close(fd);
	    execl("/tvbin/myworld", "/tvbin/myworld", NULL); exit(1);
    default: return;
  }
}
#endif


/* On exit, close the various input and output devices,
 * and reset the keyboard to its usual mode.
 */
void Exit(int how)
{
  if (exit_things & EXIT_CLOSEAVDEVS) { close(audio_fd); close(video_fd); }
  if (exit_things & EXIT_FIXKEYBOARD) tcsetattr(0, TCSANOW, &oldtios);

#ifdef TIVO
  if (exit_things & EXIT_CLOSEREMDEV) close(remote_fd);
  if (do_osd) { ClearOSD(); DrawOSD(); FreeTextOSD(); }
  if ((how==0) && (exit_things & EXIT_MYWORLD)) start_myworld();
#endif

  printf("\n\n");
  close(0);
  exit(how);
}



/* open a socket to a tcp remote host with the specified port */
int 
open_socket_to_tyserver(char *inhost, int port)
{
  int type = SOCK_STREAM;
  struct sockaddr_in sock_out;
  int res;
  int nodelay = 1;
  int window = 65535;
  char *host, *numptr, *numend;
  unsigned int raw_address=0;
  int one_part;

  host= strdup(inhost); if (host==NULL) return(-1);

  res = socket(PF_INET, type, 0);
  if (res == -1) { free(host); return (-1); }

#ifdef CAN_DO_NAME_LOOKUPS
  /* Can someone tell me how to get /etc/hosts working on the TiVo? */
  hp = gethostbyname(host);
  if (!hp) {
    fprintf(stderr, "unknown host: %s\n", host);
    return (-1);
  }
  memcpy(&sock_out.sin_addr, hp->h_addr, hp->h_length);
#else

  /* Walk the IP address and convert into a 32-bit integer */
  numptr=host; numend= strchr(host, '.');
  if (numend==NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost); 
    free(host); return(-1);
  }
  *numend='\0'; one_part=atoi(numptr);
  if ((one_part<0) || (one_part>255)) {
    fprintf(stderr, "%s not an IP address\n", inhost); 
    free(host); return(-1);
  }
  raw_address= one_part <<24;

  numptr=++numend; numend= strchr(numptr, '.');
  if (numptr==NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost); 
    free(host); return(-1);
  }
  *numend='\0'; one_part=atoi(numptr);
  if ((one_part<0) || (one_part>255)) {
    fprintf(stderr, "%s not an IP address\n", inhost); 
    free(host); return(-1);
  }
  raw_address+= (one_part <<16);
  
  numptr=++numend; numend= strchr(numptr, '.');
  if (numptr==NULL) {
    fprintf(stderr, "%s not an IP address\n", inhost); 
    free(host); return(-1);
  }
  *numend='\0'; one_part=atoi(numptr);
  if ((one_part<0) || (one_part>255)) {
    fprintf(stderr, "%s not an IP address\n", inhost); 
    free(host); return(-1);
  }
  raw_address+= (one_part <<8);
  numptr=++numend; one_part=atoi(numptr);
  if ((one_part<0) || (one_part>255)) {
    fprintf(stderr, "%s not an IP address\n", inhost); 
    free(host); return(-1);
  }
  raw_address+= one_part;
  sock_out.sin_addr.s_addr= htonl(raw_address);
#endif

  sock_out.sin_port = htons(port);
  sock_out.sin_family = PF_INET;

  if (connect(res, (struct sockaddr *)&sock_out, sizeof(sock_out))) {
    close(res);
    fprintf(stderr, "failed to connect to %s\n", inhost);
    free(host); return(-1);
  }
  setsockopt(res, SOL_SOCKET, SO_RCVBUF, &window, sizeof(int));
  setsockopt(res, IPPROTO_TCP, TCP_NODELAY, &nodelay, sizeof(int));

  free(host); return(res);
}


#ifdef TIVO
/*
 * Set the process on the TiVo to a high scheduling priority, see
 * http://www.linux.com/develop/man/2/sched_setscheduler/
 */
void 
set_high_process_priority(void)
{
  struct sched_param p;
  int err;

  p.sched_priority = 90;
  err = sched_setscheduler(0, SCHED_FIFO, &p);
}

/* Open the remote control device for reading */
void open_remote_device(void)
{
 remote_fd= open("/dev/ttyS0", O_RDONLY);
 set_nonblocking(remote_fd);
 exit_things |= EXIT_CLOSEREMDEV;
}
#endif


#ifndef TIVO
/* Set a fd into nonblocking mode */
void
set_nonblocking(int fd)
{
  int val;

  if ((val = fcntl(fd, F_GETFL, 0)) == -1) return;
  if (!(val & O_NONBLOCK)) {
    val |= O_NONBLOCK;
    fcntl(fd, F_SETFL, val);
  }
}
#endif


/* On a non-TiVo, we need to start up external processes to
 * play a stream back. We use popen and then get the
 * file descriptors for the two pipes.
 */
void
openpipes(char *audio_prog, char *video_prog)
{
  audio_f = popen(audio_prog, "w");
  if (!audio_f) {
    fprintf(stderr, "Failed to start audio program %s\n", audio_prog);
    Exit(1);
  }
  video_f = popen(video_prog, "w");
  if (!video_f) {
    fprintf(stderr, "Failed to start video program %s\n", video_prog);
    Exit(1);
  }
  audio_fd = fileno(audio_f);
  video_fd = fileno(video_f);
  exit_things |= EXIT_CLOSEAVDEVS;
}

/* On the TiVo, we can simply open up the video and audio device files */
void
opendevs(void)
{
  video_fd = open("/dev/mpeg0v", O_WRONLY);
  audio_fd = open("/dev/mpeg0a", O_WRONLY);
  exit_things |= EXIT_CLOSEAVDEVS;
}


/* Read a specific chunk from the input stream.
 * Return 1 if ok, 0 if failure
 */
int
read_a_chunk(int fd, int chunknum)
{
  char *bptr = buf;
  int numread = 0;
  int numleft = CHUNK_SIZE;

  switch (inputtype) {
#ifdef TIVO
  case INPUT_FSID:
    numread = mfs_fsid_pread(fd, buf, ((u64) chunknum) << 17, CHUNK_SIZE);
    if (numread != CHUNK_SIZE) {
      fprintf(stderr, "Failure reading chunk %d\r\n", chunknum);
      Exit(1);
    }
    return (1);
#endif
  case INPUT_STDIN:
  case INPUT_FILE:
  case INPUT_TYSERVER:
    /* We only seek if we can. This means that the forward/backward
     * commands are NOT going to work for input on stdin.
     */
    if (seeking_ok) lseek(fd, CHUNK_SIZE * chunknum, SEEK_SET);

    /* Loop until we do get CHUNK_SIZE bytes of data */
    while (numleft > 0) {
      numread = read(fd, bptr, numleft);
      if (numread < 1) {
	fprintf(stderr, "Failure reading chunk %d\r\n", chunknum);
	Exit(1);
      }
      numleft -= numread;
      bptr += numread;
    }
    return (1);
  default:
    return (0);
  }
}


#ifdef TIVO
/*
 * The TiVo remote seems to send 6 bytes of keystoke down
 * data followed by 6 bytes of up data. The down data always
 * starts with the bytes 179 122 30, then 3 bytes describing
 * the button pressed. The keystroke up data is always the
 * 6 bytes: 244 0 0 0 0 0. Here is the list as far as I can tell:
 *
 * 179 122 30 127 127 0  244 0 0 0 0 0 	TiVo Central
 * 179 122 30 59 127 1   244 0 0 0 0 0 	Live TV
 * 179 122 30 47 127 1   244 0 0 0 0 0 	Up Arrow
 * 179 122 30 35 127 1   244 0 0 0 0 0 	<--
 * 179 122 30 39 127 1   244 0 0 0 0 0 	Down Arrow
 * 179 122 30 43 127 1   244 0 0 0 0 0	-->
 * 179 122 30 31 127 1   244 0 0 0 0 0 	Thumbs Down
 * 179 122 30 27 127 1   244 0 0 0 0 0 	Select
 * 179 122 30 23 127 1   244 0 0 0 0 0 	Thumbs Up
 * 179 122 30 127 126 2  244 0 0 0 0 0 	Record
 * 179 122 30 123 126 2  244 0 0 0 0 0	Play
 * 179 122 30 15 127 1   244 0 0 0 0 0 	Volume Up
 * 179 122 30 11 127 1   244 0 0 0 0 0 	Volume Down
 * 179 122 30 7 127 1    244 0 0 0 0 0 	Channel Up
 * 179 122 30 3 127 1    244 0 0 0 0 0	Channel Down
 * 179 122 30 119 126 2  244 0 0 0 0 0 	Back
 * 179 122 30 115 126 2  244 0 0 0 0 0 	Pause
 * 179 122 30 111 126 2  244 0 0 0 0 0 	Forward
 * 179 122 30 103 126 2  244 0 0 0 0 0	Go back 30 seconds
 * 179 122 30 107 126 2  244 0 0 0 0 0 	Slow
 * 179 122 30 99 126 2   244 0 0 0 0 0 	End
 * 179 122 30 95 126 2   244 0 0 0 0 0 	1
 * 179 122 30 91 126 2   244 0 0 0 0 0 	2
 * 179 122 30 87 126 2   244 0 0 0 0 0 	3
 * 179 122 30 83 126 2   244 0 0 0 0 0 	4
 * 179 122 30 79 126 2   244 0 0 0 0 0 	5
 * 179 122 30 75 126 2   244 0 0 0 0 0	6
 * 179 122 30 71 126 2   244 0 0 0 0 0 	7
 * 179 122 30 67 126 2   244 0 0 0 0 0 	8
 * 179 122 30 63 126 3   244 0 0 0 0 0 	9
 * 179 122 30 55 126 3   244 0 0 0 0 0 	Clear
 * 179 122 30 59 126 3   244 0 0 0 0 0 	0
 * 179 122 30 51 126 3   244 0 0 0 0 0 	Enter
 *
 *
 * Read a keystroke from the TiVo remote and return its
 * information. I don't convert to editing commands here
 * so that the buttons can have different meanings in
 * different modes.
 */
int
read_a_remote_command(void)
{
  char ch;
  int ret, i, thismuch, cmd;

  while (1) {
    ret=read(remote_fd, &ch, 1);
    if (ret==-1) { return(REM_NONE); }
    if (ch!=30) continue;		/* Loop until just before key data */

    ret=read(remote_fd, &ch, 1);	/* Get 1st key byte */
    if (ret==-1) { return(REM_NONE); }

    thismuch=8;				/* Assume 8 bytes more after key id */
    switch(ch) {		       	/* Work out key id */
      case 3: cmd=  REM_CHAN_DOWN; break;
      case 7: cmd=  REM_CHAN_UP; break;
      case 11: cmd= REM_VOL_DOWN; break;
      case 15: cmd= REM_VOL_UP; break;
      case 23: cmd= REM_THUMBS_UP; break;
      case 27: cmd= REM_SELECT; break;
      case 31: cmd= REM_THUMBS_DOWN; break;
      case 35: cmd= REM_LEFT_ARROW; break;
      case 39: cmd= REM_DOWN_ARROW; break;
      case 43: cmd= REM_RIGHT_ARROW; break;
      case 47: cmd= REM_UP_ARROW; break;
      case 51: cmd= REM_ENTER; break;
      case 55: cmd= REM_CLEAR; break;
      case 63: cmd= REM_9; break;
      case 67: cmd= REM_8; break;
      case 71: cmd= REM_7; break;
      case 75: cmd= REM_6; break;
      case 79: cmd= REM_5; break;
      case 83: cmd= REM_4; break;
      case 87: cmd= REM_3; break;
      case 91: cmd= REM_2; break;
      case 95: cmd= REM_1; break;
      case 99: cmd= REM_END; break;
      case 103: cmd= REM_BACK30; break;
      case 107: cmd= REM_SLOW; break;
      case 111: cmd= REM_FORWARD; break;
      case 115: cmd= REM_PAUSE; break;
      case 119: cmd= REM_BACK; break;
      case 123: cmd= REM_PLAY; break;

      case 59:
	thismuch=7;
        ret=read(remote_fd, &ch, 1);
        if (ret==-1) { return(REM_NONE); }
	switch (ch) {
          case 126: cmd=REM_0; break;
          case 127: cmd=REM_LIVE_TV; break;
	  default:  cmd=REM_NONE;
	}
	break;

      case 127:
	thismuch=7;
        ret=read(remote_fd, &ch, 1);
        if (ret==-1) { return(REM_NONE); }
	switch (ch) {
          case 126: cmd=REM_RECORD; break;
          case 127: cmd=REM_TIVO_CENTRAL; break;
	  default:  cmd=REM_NONE;
	}
	break;

      default: cmd=REM_NONE;
    }

    /* Now read in the key up data to clean up */
    for (i=0;i<thismuch;i++) {
      ret=read(remote_fd, &ch, 1);
      if (ret==-1) break;
    }
    return(cmd);
  }					/* end of forever loop */
}
#endif


/* Obtain a command from the user and return it.
 * Commands can come in from stdin or the TiVo remote.
 */
int
read_a_command(void)
{
  char ch;

#ifdef TIVO
  int key;

  /* See if there is a keystroke from the TiVo remote control
   * before we try the telnet keyboard.
   */
  key= read_a_remote_command();

  switch(key) {
    case REM_TIVO_CENTRAL: return(CMD_QUIT);
    case REM_RECORD: return(CMD_RECORD);
    case REM_SELECT: return(CMD_PLAYITALL);
    case REM_PLAY: return(CMD_NOPAUSE);
    case REM_BACK: return(CMD_BACKFRAME);
    case REM_PAUSE: return(CMD_PAUSE);
    case REM_FORWARD: return(CMD_FWDFRAME);
    case REM_BACK30: return(CMD_START);
    case REM_END: return(CMD_END);
    case REM_1: return(CMD_FASTF1);
    case REM_2: return(CMD_FASTF2);
    case REM_3: return(CMD_FASTF3);
    case REM_4: return(CMD_FASTF4);
    case REM_5: return(CMD_FASTF5);
    case REM_6: return(CMD_FASTF6);
    case REM_7: return(CMD_FASTB2);
    case REM_8: return(CMD_FASTB3);
    case REM_9: return(CMD_FASTB4);
    case REM_CLEAR: return(CMD_CSTART);
    case REM_ENTER: return(CMD_CEND);
    default: break;
  }
#endif

  if (inputtype == INPUT_STDIN) return (CMD_NONE); /* No cmds, only stream */

  if (read(0, &ch, 1) != 1) return (CMD_NONE);	/* No kybrd data at present */

  switch (ch) {
  case ' ': return (CMD_PAUSE); break;
  case 'f': return (CMD_FWDFRAME); break;
  case 'b': return (CMD_BACKFRAME); break;
  case  3 :					/* ctrl-C and q */
  case 'Q':
  case 'q': return (CMD_QUIT); break;
  case 's':
  case 'c': return (CMD_CSTART); break;
  case 'e': return (CMD_CEND); break;
  case 'R': return (CMD_RECORD); break;
  case 'P': return (CMD_PLAYITALL); break;
  case '1': return (CMD_FASTF1); break;
  case '2': return (CMD_FASTF2); break;
  case '3': return (CMD_FASTF3); break;
  case '4': return (CMD_FASTF4); break;
  case '5': return (CMD_FASTF5); break;
  case '6': return (CMD_FASTF6); break;
  case '!': return (CMD_FASTB1); break;
  case '@': return (CMD_FASTB2); break;
  case '#': return (CMD_FASTB3); break;
  case '$': return (CMD_FASTB4); break;
  case '%': return (CMD_FASTB5); break;
  case '^': return (CMD_FASTB6); break;
  case '0': return (CMD_END); break;
  case ')': return (CMD_START); break;
  case 'k': return (CMD_KILLCUT); break;
  case 'K': return (CMD_KILLALL); break;
  case 'l': return (CMD_LISTCUTS); break;
  }
  return (CMD_NONE);
}


/* Now a misnomer, this routine plays a chunk back. */
static void
parse_chunk(void)
{
  int num_recs = buf[0];
  int i;
  unsigned char *p;
  int ofs = 0;

  /*
   * Each 128K chunk starts with N 16 byte records that
   * tell you what sorts of things are in the chunk.
   */
  p = &buf[4];
  ofs = 4 + num_recs * 16;

  for (i = 0; i < num_recs; i++) {
    unsigned size = (p[0] << 8 | p[1]) << 4 | (p[2] >> 4);
    unsigned type = p[3];

    if (type == 0xe0) {
#ifdef TIVO
      write(video_fd, &buf[ofs], size);
#else
      if (size != 0x10) {
	write(video_fd, &buf[ofs], size);
	if (verbose) {
	  fprintf(stderr, "video at 0x%x of size 0x%x\r\n", ofs, size);
	}
      }
#endif
      ofs += size;
      p += 16;
      continue;
    }
    if (type == 0xc0) {
#ifdef TIVO
      write(audio_fd, &buf[ofs], size);
#else
      if (size != 0x10) {
	write(audio_fd, &buf[ofs], size);
	if (verbose) {
	  fprintf(stderr, "audio at 0x%x of size 0x%x\r\n", ofs, size);
	}
      }
#endif
      ofs += size;
      p += 16;
      continue;
    }
    p += 16;
  }
}


/* Given the input argument from the user,
 * work out which one it is and open it.
 * Returns a file descriptor to the input,
 * or a fsid if the input is on MFS.
 */
int
open_input(char *inputsource)
{
#ifdef TIVO
  int fsid;
#endif
  int fd;
  struct stat statbuf;
  char ch, *tyfile;
  char result[4];

  /* Use stdin if the input source is "-" */
  if (!strcmp(inputsource, "-")) {
    seeking_ok = 0; totalchunks = CHUNKSUNKNOWN;
    inputtype = INPUT_STDIN; return (0);
  }

  /* If the input source has a colon, get it from the tyserver */

  if ((tyfile=strchr(inputsource, ':'))!=NULL) {
    *tyfile='\0'; tyfile++;
    fd= open_socket_to_tyserver(inputsource, 1150);
    if (fd==-1) Exit(1);
   
    /* Now send a PLAY command, and get a 200 reply */ 
    write(fd, "PLAY ", 5);
    write(fd, tyfile, strlen(tyfile));
    write(fd, "\r\n", 2);
    read(fd, result, 3); result[3]='\0';
    if (strcmp(result, "200")) {      /* Did not get ok message */
      close(fd); fprintf(stderr, "Couldn't play from tyserver\n");
      Exit(1);
    }

    /* Otherwise, read up to \n from tyserver before we get the stream */
    while (1) { read(fd, &ch, 1); if (ch=='\n') break; }
    seeking_ok = 0; totalchunks = CHUNKSUNKNOWN;
    inputtype = INPUT_TYSERVER; return(fd);
  }

  /* If the input source is numeric, treat it as a fsid */
#ifdef TIVO
  fsid = atoi(inputsource);
  if (fsid != 0) {
    mfs_init();
    totalchunks = mfs_fsid_size(fsid) >> 17;
    fprintf(stderr, "fsid %d has %lld chunks\n",
	    fsid, mfs_fsid_size(fsid) >> 17);
    seeking_ok = 0; inputtype = INPUT_FSID; return (fsid);
  }
#endif

  /* None of the above, treat it as a Unix filename */
  fd = open(inputsource, O_RDONLY);
  if (fd == -1) {
    fprintf(stderr, "Cannot open input file %s\n", inputsource);
    Exit(1);
  }
  if (stat(inputsource, &statbuf) == -1) {
    fprintf(stderr, "Cannot determine size of %s\n", inputsource);
    Exit(1);
  }
  totalchunks = statbuf.st_size / CHUNK_SIZE;
  seeking_ok = 1; inputtype = INPUT_FILE;
  return (fd);
}

/* Playback a program to the audio/video devices,
 * removing the cut chunks, and preventing editing
 * from occurring. Also give delay before starting.
 * fd is stream input, ncuts is # of cuts in list.
 */
void 
playback_program(int fd, int ncuts)
{
  int chk, send_it, i;

  fprintf(stderr, "\nStarting to playback in\n");
#ifdef TIVO
  if (do_osd) {
    osd_x=1, osd_y=15;
    ClearOSD();
    DrawString(&osd_x, &osd_y, "Playback starting ...", 4, 1);
    DrawOSD(); sleep(1);
  }
#endif

  for (i = 10; i > 0; i--) {
    fprintf(stderr, "%d seconds ...\n", i);
#ifdef TIVO
    if (do_osd) {
      osd_x=1, osd_y=15;
      ClearOSD();
      if (i>3) {
        snprintf(osd_string, 79, "%d seconds  ", i);
        DrawString(&osd_x, &osd_y, osd_string, 4, 1);
      }
      DrawOSD(); sleep(1);
    }
#endif
    sleep(1);
  }

#ifdef TIVO
  if (do_osd) { ClearOSD(); DrawOSD(); }
#endif

  for (chk = 1; chk < totalchunks; chk++) {
				/* Work out if we should play this chunk */
    send_it = 1;
    for (i = 0; i < ncuts; i++) {
      if ((cutstart[i] == 0) || (cutend[i] == 0)) continue;
      if ((chk >= cutstart[i]) && (chk <= cutend[i])) { send_it = 0; break; }
    }
    if (send_it == 0) {
      if (verbose) fprintf(stderr, "Skipping %d\n", chk);
      continue;
    }
    read_a_chunk(fd, chk);
    if (allow_playback) parse_chunk();
  }
  fprintf(stderr, "Playback done...\n");
}

/* Export a program to some output file, removing the
 * cut chunks, and preventing editing from occurring.
 * fd is stream input, ncuts is # of cuts in list.
 */
void
record_program(char *outfile, int fd, int ncuts)
{
  int numleft, numsent;
  unsigned char *bptr;
  int chk, send_it, i;
  int outfd;
  char *tyfile;
  char ch, result[4];

  if (outfile==NULL) { fprintf(stderr, "No output file was named\n"); return; }

		/* - means standard output */
  if (!strcmp(outfile, "-")) { outfd=1; }

		/* a name with a colon means to send it to the tyserver */
  else if ((tyfile=strchr(outfile, ':'))!=NULL) {
      *tyfile='\0'; tyfile++;
      outfd= open_socket_to_tyserver(outfile, 1150);
      if (outfd==-1) return;

      /* Now send a SAVE command, and get a 200 reply */
      write(outfd, "SAVE ", 5);
      write(outfd, tyfile, strlen(tyfile));
      write(outfd, "\r\n", 2);
      read(outfd, result, 3); result[3]='\0';
      if (strcmp(result, "200")) {	/* Did not get ok message */
	close(outfd); fprintf(stderr, "Couldn't record to tyserver\n");
	return;
      }

      /* Otherwise, read up to \n from tyserver before we send */
      while (1) { read(outfd, &ch, 1); if (ch=='\n') break; }
  }

			/* Otherwise, assume a Unix filename */
  else {
    outfd=open(outfile, O_WRONLY|O_CREAT|O_TRUNC, 0644);
    if (outfd==-1) {
      fprintf(stderr, "Cannot write to %s\n", outfile);
      return;
    }
  }

  fprintf(stderr, "Starting to record ...\n");
#ifdef TIVO
  if (do_osd) {
    osd_x=1, osd_y=15;
    ClearOSD();
    DrawString(&osd_x, &osd_y, "Recording ...", 4, 1);
    DrawOSD();
  }
#endif
  for (chk = 0; chk < totalchunks; chk++) {
				/* Work out if we should play this chunk */
    send_it = 1;
    for (i = 0; i < ncuts; i++) {
      if ((cutstart[i] == 0) || (cutend[i] == 0)) continue;
      if ((chk >= cutstart[i]) && (chk <= cutend[i])) { send_it = 0; break; }
    }
    if (send_it == 0) {
      if (verbose) fprintf(stderr, "Skipping %d\n", chk);
      continue;
    }
    read_a_chunk(fd, chk);
    if (chk && ((chk % 100) == 0)) {
      fprintf(stderr, "Chunk %d\n", chk);
#ifdef TIVO
      if (do_osd) {
        snprintf(osd_string, 79, "Chunk %d", chk);
        osd_x=1, osd_y=15;
        ClearOSD();
        DrawString(&osd_x, &osd_y, osd_string, 4, 1);
        DrawOSD();
      }
#endif
    }

    numleft = CHUNK_SIZE;
    bptr = buf;
    while (numleft > 0) {
      numsent = write(outfd, bptr, numleft);
      if (numsent < 1) { fprintf(stderr, "Failure saving chunk\n"); Exit(1); }
      numleft -= numsent;
      bptr += numsent;
    }
    if (verbose) fprintf(stderr, "Chunk %d\n", chk);
  }
  fprintf(stderr, "Recording done\n");
#ifdef TIVO
  if (do_osd) { ClearOSD(); DrawOSD(); }
#endif
}

/* Print out usage information on the application & exit */
static void
usage(void)
{
  fprintf(stderr, "playitsam [options] inputfile [outputfile]\n");
  fprintf(stderr, "options:\n");
  fprintf(stderr, "\t -v verbose\n ");
  fprintf(stderr, "\t -n don't play the stream to a/v device\n ");
  fprintf(stderr, "\t -p only play the stream, disallow editing\n ");
#ifdef TIVO
  fprintf(stderr, "\t -M restart myworld when playitsam exits\n");
  fprintf(stderr, "\t -D don't give on-screen user prompting\n");
#endif
  fprintf(stderr, "\t -h print this help information\n\n");
  fprintf(stderr, "input source is one of the following:\n");
  fprintf(stderr, "\t - means standard input (no skipping fwd/back)\n");
#ifdef TIVO
  fprintf(stderr, "\t NNNN means the fsid on MFS with that number\n");
#endif
  fprintf(stderr, "\t xxx.xxx.xxx.xxx:name means a file from a tyserver\n");
  fprintf(stderr, "\t anything else means a Unix filename\n\n");
  fprintf(stderr, "output source is one of the following:\n");
  fprintf(stderr, "\t - means standard output\n");
  fprintf(stderr, "\t xxx.xxx.xxx.xxx:name means a file to a tyserver\n");
  fprintf(stderr, "\t anything else means to a Unix filename\n\n");
  Exit(1);
}


int
main(int argc, char *argv[])
{
  char *fname;			/* Input file name */
  char *outfname=NULL;		/* Output (recording) file name */
  char *argflags;
  int fd;
  extern char *optarg;
  extern int optind;
  int c, cmd;
  int paused = 0;		/* Start in non-paused mode */
  int cutnum = 1;		/* and assuming user wants to cut chunk 1 on */

#ifdef TIVO
  argflags= "vhnpMD";
#else
  argflags= "vhnp";
#endif

  /* Get the various flags to the program */
  while ((c = getopt(argc, argv, argflags)) != -1) {
    switch (c) {
    case 'v': verbose++; break;

    case 'n': if (allow_playback==2) {
                fprintf(stderr, "Tell me, why did you try -n and -p?\n");
              }
	      allow_playback = 0; break;

    case 'p': if (allow_playback==0) {
		fprintf(stderr, "Tell me, why did you try -n and -p?\n");
	      }
	      allow_playback = 2; break;

    case 'M': exit_things|= EXIT_MYWORLD; break;

    case 'D': do_osd=0; break;

    case 'h':
    default: usage();
    }
  }

  /* and skip over to the input argument */
  argc -= optind;
  argv += optind;
  if (argc < 1) { usage(); }

#ifdef TIVO
  if (allow_playback) {
    /* We need to kill both EventSwitcher and MfsDaemon */
    c= pid_fromname("E ventSwitcher"); if (c>0) {
      kill(c, SIGTERM); fprintf(stderr, "Killed EventSwitcher pid %d\n", c);
    }
    c= pid_fromname("MfsDaemon"); if (c>0) {
      kill(c, SIGTERM); fprintf(stderr, "Killed MfsDaemon pid %d\n", c);
    }
  }
#endif

  /* Set the keyboard to raw mode & non-blocking,
   * so we can get single character commands
   */
  if (allow_playback!=2) {
    tcgetattr(0, &oldtios);
    memcpy(&newtios, &oldtios, sizeof(oldtios));
    set_nonblocking(0);
    cfmakeraw(&newtios);
    tcsetattr(0, TCSANOW, &newtios);
    exit_things |= EXIT_FIXKEYBOARD;
  }

  /* Try to open the input up */
  fname = argv[0]; outfname = argv[1];
  fd = open_input(fname);
  if (fd==-1) { fprintf(stderr, "Can't open input file %s\n",fname); Exit(1); }

  /* Open the output devices if we are going to play things back */
  if (allow_playback) {
#ifdef TIVO
    opendevs();
    open_remote_device();
    set_high_process_priority();
#else
    openpipes(AUDIO_PROG, VIDEO_PROG);
#endif
  }

#ifdef TIVO
 if (do_osd) { SetupTextOSD(); ClearOSD(); DrawOSD(); }
#endif

  /* Clear the list of cuts */
  for (c = 0; c < MAXCUTS; c++) cutstart[c] = cutend[c] = 0;

  /* and mark to cut from chunk 1 */
  cutstart[cutnum] = 1;

  /* discard the first chunk - it often contains garbage */
  if (inputtype != INPUT_FSID) { read_a_chunk(fd, 0); }
  chunk = 1;


  /* If they only want playback, give it to them */
  if (allow_playback==2) { playback_program(fd, 0); Exit(0); }

  /* Otherwise, loop reading chunks from stream input, and dealing
   * with user commands. Play back the stream if allowed.
   */
  while (1) {
    if (chunk < totalchunks) {
      read_a_chunk(fd, chunk);
      if (allow_playback) parse_chunk();
    }

here:
    fprintf(stderr, "\rChunk %d", chunk);
    if (totalchunks != CHUNKSUNKNOWN)
		fprintf(stderr, " out of %d", totalchunks);
    fprintf(stderr, "       ");
#ifdef TIVO
    if (do_osd && paused && (osd_chunk!=chunk)) {
      snprintf(osd_string, 79, "Chunk %d", chunk);
      osd_x=1, osd_y=15;
      ClearOSD();
      DrawString(&osd_x, &osd_y, osd_string, 4, 1);
      DrawOSD();
      osd_dirty=1; osd_chunk=chunk;
    }
#endif

    cmd = read_a_command();
    switch (cmd) {
    case CMD_NONE: break;
    case CMD_PAUSE: paused = 1 - paused; break;
    case CMD_NOPAUSE: paused = 0; break;
    case CMD_END: chunk = totalchunks - 1; paused = 1; break;
    case CMD_START: chunk = 1; break;
    case CMD_FWDFRAME: chunk++; break;
    case CMD_BACKFRAME: chunk--; break;
    case CMD_FASTF1: chunk += 5; break;
    case CMD_FASTF2: chunk += 25; break;
    case CMD_FASTF3: chunk += 125; break;
    case CMD_FASTF4: chunk += 500; break;
    case CMD_FASTF5: chunk += 2500; break;
    case CMD_FASTF6: chunk += 12500; break;
    case CMD_FASTB1: chunk -= 5; break;
    case CMD_FASTB2: chunk -= 25; break;
    case CMD_FASTB3: chunk -= 125; break;
    case CMD_FASTB4: chunk -= 500; break;
    case CMD_FASTB5: chunk -= 2500; break;
    case CMD_FASTB6: chunk -= 12500; break;

   				/* On quit, reset the keyboard to cooked */
    case CMD_QUIT: Exit(0);

    case CMD_CSTART:
      if (cutnum >= MAXCUTS) break;
      cutstart[cutnum] = chunk;
      fprintf(stderr, "\r\nCut starting at chunk %d\r\n", chunk);
#ifdef TIVO
      if (do_osd && paused) {
        snprintf(osd_string, 79, "Cut starting at %d", chunk);
        osd_x=1, osd_y=15;
        ClearOSD();
        DrawString(&osd_x, &osd_y, osd_string, 4, 1);
        DrawOSD();
        osd_dirty=1;
      }
#endif
      break;

    case CMD_CEND:
      if (cutnum >= MAXCUTS) {
	fprintf(stderr, "\r\nNo more than %d cuts available!\r\n", MAXCUTS);
	break;
      }
      if (cutstart[cutnum] == 0) {
	fprintf(stderr, "\r\nThere is no cut at present for you to end\r\n");
	break;			/* No cut */
      }
      if (chunk <= cutstart[cutnum]) {
	fprintf(stderr,
		"\r\nYou can't cut at chunk %d, as it's before chunk %d\r\n",
		chunk, cutstart[cutnum]);
	break;			/* No cut */
      }
      cutend[cutnum] = chunk;
      fprintf(stderr, "\r\nNew cut from %d to %d\r\n",
	      cutstart[cutnum], cutend[cutnum]);
#ifdef TIVO
      if (do_osd && paused) {
        snprintf(osd_string, 79, "New cut %d to %d",
	      			cutstart[cutnum], cutend[cutnum]);
        osd_x=1, osd_y=15;
        ClearOSD();
        DrawString(&osd_x, &osd_y, osd_string, 4, 1);
        DrawOSD();
        osd_dirty=1;
      }
#endif
      cutnum++;
      break;

    case CMD_RECORD:
      tcsetattr(0, TCSANOW, &oldtios);
      record_program(outfname, fd, cutnum);
      Exit(0);

    case CMD_PLAYITALL:
      if (allow_playback == 0) {
	fprintf(stderr, "\r\nYou can't playback with the -n flag\r\n");
	break;
      }
      tcsetattr(0, TCSANOW, &oldtios);
      playback_program(fd, cutnum);
      tcsetattr(0, TCSANOW, &newtios);
      chunk = totalchunks - 1;
      paused = 1;
      break;

    case CMD_KILLCUT:
      if (cutstart[cutnum] == 0)
	fprintf(stderr, "\r\nThere is no cut at present for you to end\r\n");
      else {
	fprintf(stderr, "\r\nKilling cut starting at chunk %d\r\n",
		cutstart[cutnum]);
	cutstart[cutnum] = 0;
      }
      break;

    case CMD_KILLALL:
      fprintf(stderr, "\r\nKilling all cuts\r\n");
      for (c = 0; c < MAXCUTS; c++) cutstart[c] = cutend[c] = 0;
      break;

    case CMD_LISTCUTS:
      fprintf(stderr, "\r\nHere is the list of cuts:\r\n");
      for (c = 0; c < MAXCUTS; c++) {
	if (cutstart[c] != 0) {
	  if (cutend[c] != 0)
	    fprintf(stderr, "  cut from chunk %d to %d\r\n",
		    cutstart[c], cutend[c]);
	  else
	    fprintf(stderr, "  cut opened with start at chunk %d\r\n",
		    cutstart[c]);
	}
      }
      break;
    }

    /* Make sure that the chunk falls into acceptable limits */
    if (chunk < 1) chunk = 1;
    if (chunk >= totalchunks) { chunk = totalchunks - 1; paused = 1; }

    /* For certain command during a pause, don't play another chunk */
    if (paused)
      switch (cmd) {
      case CMD_CSTART:
      case CMD_CEND:
      case CMD_PAUSE:
      case CMD_NONE: usleep(100000); goto here;		/* Yuk, a goto! */

    /* If not paused, go on to the next chunk */
    } else {
      chunk++;
#ifdef TIVO
      if (osd_dirty) {
        ClearOSD();
        DrawOSD();
	osd_dirty=0; osd_chunk=0;
      }
#endif
    }
  }

  Exit(0); return(0);
}
